package ioz

import (
	"context"
	"crypto/ed25519"
	"crypto/sha256"
	"encoding/base64"
	"fmt"
	"io"
	"log/slog"
	"os"
	"sync"

	"google.golang.org/grpc/mem"
)

const (
	// hmacSHA256MinKeySize is the minimum HMAC SHA256 key size that is considered secure.
	// See https://www.rfc-editor.org/rfc/rfc2104#section-3 and https://en.wikipedia.org/wiki/HMAC.
	hmacSHA256MinKeySize = sha256.Size
	// hmacSHA256MaxKeySize is the maximum HMAC SHA256 key size that is not hashed into 32 bytes.
	hmacSHA256MaxKeySize = sha256.BlockSize
	// hmacSHA3_512MinKeySize is the minimum HMAC SHA3 512 key size that is considered secure.
	// See https://www.rfc-editor.org/rfc/rfc2104#section-3 and https://en.wikipedia.org/wiki/HMAC.
	hmacSHA3_512MinKeySize = 64
	// hmacSHA3_512MaxKeySize is the maximum HMAC SHA3 512 key size that is not hashed into 64 bytes.
	hmacSHA3_512MaxKeySize = 72
	// aes256KeyLength is the expected key length for the Workflow data encryption.
	// crypto/aes forces a 32 key int AES-265.
	aes256KeyLength = 32
)

// DiscardData discards up to 8KiB of data from r.
// Function can be used to drain HTTP client response body.
// See https://pkg.go.dev/net/http#Response.
func DiscardData(r io.Reader) error {
	_, err := io.Copy(io.Discard, io.LimitReader(r, 8*1024))
	if err != nil {
		return fmt.Errorf("failed to read data: %w", err)
	}
	return nil
}

func LoadSHA256Base64Secret(log *slog.Logger, filename string) ([]byte, error) {
	return LoadBase64Secret(log, filename, hmacSHA256MinKeySize, hmacSHA256MaxKeySize)
}

func LoadSHA3_512Base64Secret(log *slog.Logger, filename string) ([]byte, error) {
	return LoadBase64Secret(log, filename, hmacSHA3_512MinKeySize, hmacSHA3_512MaxKeySize)
}

func LoadEd25519Base64PublicKey(log *slog.Logger, filename string) ([]byte, error) {
	return LoadBase64Secret(log, filename, ed25519.PublicKeySize, ed25519.PublicKeySize)
}

func LoadBase64Secret(log *slog.Logger, filename string, minLen, maxLen int) ([]byte, error) {
	encodedAuthSecret, err := os.ReadFile(filename) //nolint: gosec
	if err != nil {
		return nil, fmt.Errorf("read file: %w", err)
	}
	decodedAuthSecret := make([]byte, base64.StdEncoding.DecodedLen(len(encodedAuthSecret)))

	n, err := base64.StdEncoding.Decode(decodedAuthSecret, encodedAuthSecret)
	if err != nil {
		return nil, fmt.Errorf("decoding: %w", err)
	}

	if n < minLen || n > maxLen {
		// TODO make this an error in 18.0, remove logging
		// TODO https://gitlab.com/gitlab-org/cluster-integration/gitlab-agent/-/issues/617
		log.Warn(fmt.Sprintf("secret in file %s is %d bytes long; allowed length is from %d to %d bytes", filename, n, minLen, maxLen))
		//return nil, fmt.Errorf("secret is %d bytes long; allowed length is from %d to %d bytes", n, minLen, maxLen)
	}

	return decodedAuthSecret[:n], nil
}

// ReadAllFunc fully reads the reader and passes the data to the callback.
// Callback must not retain the buffer or any portion of it.
func ReadAllFunc(r io.Reader, f func([]byte) error) error {
	bs, err := mem.ReadAll(r, mem.DefaultBufferPool())
	if err != nil {
		bs.Free()
		return fmt.Errorf("read all: %w", err)
	}
	data := bs.MaterializeToBuffer(mem.DefaultBufferPool())
	bs.Free()
	defer data.Free()
	return f(data.ReadOnlyData())
}

func ReadSecretAES256Key(path string) ([]byte, error) {
	k, err := os.ReadFile(path) //nolint: gosec
	if err != nil {
		return nil, fmt.Errorf("unable to read key file: %w", err)
	}

	if len(k) != aes256KeyLength {
		return nil, fmt.Errorf("secret is %d bytes but want %d", len(k), aes256KeyLength)
	}

	return k, nil
}

type CancelingReadCloser struct {
	io.ReadCloser
	Cancel context.CancelFunc
}

func (c *CancelingReadCloser) Close() error {
	// First close the pipe, then cancel the context.
	// Doing it the other way around can result in another goroutine calling CloseWithError() with the
	// cancellation error (if ReadCloser is a io.Pipe() used across goroutines).
	err := c.ReadCloser.Close()
	c.Cancel()
	return err
}

type OnceReadCloser struct {
	io.ReadCloser
	once     sync.Once
	closeErr error
}

func (oc *OnceReadCloser) Close() error {
	oc.once.Do(oc.close)
	return oc.closeErr
}

func (oc *OnceReadCloser) close() {
	oc.closeErr = oc.ReadCloser.Close()
}
